# -*- coding: utf-8 -*-
"""
:Author: ChenXiaolei
:Date: 2020-04-16 21:32:43
:LastEditTime: 2021-02-03 19:35:45
:LastEditors: ChenXiaolei
:Description: 日志帮助类
"""

import logging
import logging.handlers
import time
import os
import json
import socket
import platform

from seven_framework.mysql import *
from seven_framework.json import *


class Logger:
    """
    指定保存日志的文件路径，日志级别，以及调用文件 将日志存入到指定的文件中
    级别优先级:NOTSET < DEBUG < INFO < WARNING < ERROR < CRITICAL
    """
    def __init__(self,
                 log_file_name,
                 log_level,
                 logger,
                 host_ip="",
                 project_name=None,
                 storage=None):
        """
        :Description: 
        :param log_file_name: 日志存储文件路径
        :param log_level: 日志等级
        :param logger: 日志标识
        :param host_ip: 服务器IP
        :param project_name: 项目标志
        :last_editors: ChenXiaolei
        """
        if not project_name:
            if platform.system() == "Windows":
                project_name = os.getcwd().split('\\')[-1]
            else:
                project_name = os.getcwd().split('/')[-1]

        # 判断文件夹是否存在，不存在则创建
        path_list = log_file_name.split("/")
        path_log = log_file_name[0:log_file_name.find(path_list[len(path_list)
                                                                - 1])]
        if not os.path.isdir(path_log):
            os.mkdir(path_log)

        logging_level = ''

        if log_level.upper() == 'NOTSET':
            logging_level = logging.NOTSET
        elif log_level.upper() == 'DEBUG':
            logging_level = logging.DEBUG
        elif log_level.upper() == 'INFO':
            logging_level = logging.INFO
        elif log_level.upper() == 'WARNING':
            logging_level = logging.WARNING
        elif log_level.upper() == 'ERROR':
            logging_level = logging.ERROR
        elif log_level.upper() == 'CRITICAL':
            logging_level = logging.CRITICAL
        elif log_level.upper() == 'SQL':
            logging_level = logging.INFO
        elif log_level.upper() == 'HTTP':
            logging_level = logging.INFO

        # 创建一个logger
        self.logger = logging.getLogger(logger)
        self.logger.setLevel(logging_level)
        # 关闭原始日志在控制台显示(表面日志多次打印)
        self.logger.propagate = False

        # 创建handler，用于写入日志文件
        self.handler_file = logging.handlers.TimedRotatingFileHandler(
            log_file_name, 'D', 1, 10)
        # 设置 切分后日志文件名的时间格式 默认 log_file_name+"." + suffix 如果需要更改需要改logging 源码
        self.handler_file.suffix = "%Y%m%d.log"
        self.handler_file.setLevel(logging_level)

        formatter = logging.Formatter(
            JsonHelper.json_dumps({
                "record_time": "%(asctime)s",
                "level": "%(levelname)s",
                "log_msg": "%(message)s",
                "host_ip": host_ip,
                "project_name": project_name
            }))
        self.handler_file.setFormatter(formatter)

        # 创建handler，用于输出至控制台
        # 定义控制台输出handler的输出格式
        formatter = logging.Formatter(
            '[%(asctime)s][%(name)s][%(levelname)s]%(message)s')
        self.handler_console = logging.StreamHandler()  # 输出到控制台的handler
        self.handler_console.setFormatter(formatter)

        # 给logger添加handler
        if not self.logger.handlers:
            self.logger.addHandler(self.handler_file)
            self.logger.addHandler(self.handler_console)

        # 日志存储
        if storage and type(storage) == dict:
            self.handler_storage = LoggerStorageHandler(
                log_level, storage, project_name, host_ip)
            self.logger.addHandler(self.handler_storage)

    def close(self):
        self.logger.removeHandler(self.handler_file)
        self.handler_file.close()

        if hasattr(self, "handler_console"):
            self.logger.removeHandler(self.handler_console)
            self.handler_console.close()
        if hasattr(self, "handler_storage"):
            self.logger.removeHandler(self.handler_storage)
            self.handler_storage.close()

    def get_logger(self):
        return self.logger

    @classmethod
    def get_logger_by_name(self, loger_name):
        """
        :Description: 通过日志标识获取logger
        :param loger_name: 日志标识
        :return: logger
        :last_editors: ChenXiaolei
        """
        return logging.getLogger(loger_name)


class LoggerStorageHandler(logging.Handler):
    def __init__(self, level, storage, project_name="", host_ip=""):
        if "engine" not in storage and "config" not in storage:
            raise "未配置log存储介质"

        self.project_name = project_name
        self.host_ip = host_ip

        self.storage_engine = storage["engine"]

        if self.storage_engine.lower() == "mysql":
            try:
                from seven_framework.mysql import MySQLHelper
                self.mysql_db = MySQLHelper(storage["config"])
                self.mysql_db._log_sql = None

                # 获取表名
                self.table_name = "python_log_tb"

                if "level_setting" in storage:
                    for level_item in storage["level_setting"]:
                        if level_item["level"].lower() == level.lower():
                            self.table_name = level_item["key"]

                # 判断制定数据库中标是否存在
                exist_result = self.mysql_db.fetch_one_row(
                    f"SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA='{storage['config']['db']}' AND TABLE_NAME='{self.table_name}';"
                )

                if not exist_result:
                    self.mysql_db.fetch_and_commit(
                        f"CREATE TABLE `{self.table_name}` ( `id` int(11) NOT NULL AUTO_INCREMENT COMMENT '主键id', `request_code` varchar(50) DEFAULT NULL COMMENT '请求标识', `project_name` varchar(50) DEFAULT NULL COMMENT '项目名称', `level` varchar(20) DEFAULT NULL COMMENT '日志等级', `host_ip` varchar(50) DEFAULT NULL COMMENT '服务器ip', `record_time` datetime NOT NULL COMMENT '记录时间', `record_timestamp` int(11) NOT NULL COMMENT '记录时间',`log_msg` longtext NOT NULL COMMENT '日志内容', PRIMARY KEY (`id`) ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
                    )
            except:
                print("构建mysql日志环境异常:" + traceback.format_exc())
                pass
        elif self.storage_engine.lower() == "redis":
            from seven_framework.redis import RedisHelper
            try:
                self.redis_log_client = RedisHelper.redis_init(
                    config_dict=storage["config"])
                # 获取表名
                self.redis_list_key = "python_log_list"

                if "level_setting" in storage:
                    for level_item in storage["level_setting"]:
                        if level_item["level"].lower() == level.lower():
                            self.redis_list_key = level_item["key"]
                            # 可定制不通level走不通的redis连接
                            if "config" in level_item["level"]:
                                self.redis_log_client = RedisHelper.redis_init(
                                    level_item["level"]["config"])
            except:
                print("构建redis日志环境异常:" + traceback.format_exc())
                pass

        logging.Handler.__init__(self)

    def emit(self, record):
        """
        :description: logging.Handler内置事件
        :param record: 日志信息
        :return: 无
        :last_editors: ChenXiaolei
        """
        request_code = self.request_code if hasattr(self,
                                                    "request_code") else ""
        project_name = self.project_name
        level = record.levelname
        host_ip = self.host_ip
        record_time = int(record.created)
        log_msg = record.message

        if self.storage_engine == "mysql":
            try:
                self.mysql_db.fetch_and_commit(
                    f"INSERT INTO {self.table_name}(request_code,project_name,level,host_ip,record_time,record_timestamp,log_msg) VALUE('{request_code}','{project_name}','{level}','{host_ip}','{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(record_time))}',{record_time},'{pymysql.converters.escape_string(log_msg)}');"
                )
            except:
                print("日志入库异常:" + traceback.format_exc())
                pass

        elif self.storage_engine == "redis":
            try:
                log_info = {}
                log_info["request_code"] = request_code
                log_info["project_name"] = project_name
                log_info["level"] = level
                log_info["host_ip"] = host_ip
                log_info["record_time"] = record_time
                log_info["log_msg"] = log_msg
                self.redis_log_client.lpush(self.redis_list_key,
                                            JsonHelper.json_dumps(log_info))
            except:
                print("日志lpush至redis时异常:" + traceback.format_exc())
                pass